import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;


/**
 * Contains classes for working with files as streams of BitStrings
 */
public class BitStringFile
{
	/**
	 * A class for reading the bits of a file as bitstrings
	 */
	public static class Reader
	{
		private BufferedReader reader; //does the actual reading
		
		/**
		 * Creates a new Reader for the given file source
		 * @param source The file to read
		 * @throws FileNotFoundException
		 */
		public Reader(File source) throws FileNotFoundException
		{
			reader = new BufferedReader(new FileReader(source));			
		}
		
		/**
		 * Reads a single line of text from the source file. This method advances the Reader's location in the file
		 * @return The text read
		 * @throws IOException
		 */
		public String readLine() throws IOException
		{
			return reader.readLine();
		}
		
//		/**
//		 * Reads a single character (one byte) from the source file. This method advances the Reader's location in the file
//		 * @return The character read
//		 * @throws IOException
//		 */
//		public char readCharacter() throws IOException
//		{
//			
//			
//		}
		
		/**
		 * Reads the <u>remainder</u> of the source file and returns a BitString representation of that content.
		 * This method will advance the Reader's location to the end of the file.
		 * This method also closes the file.
		 * @return
		 */
		public String readBitString() throws IOException
		{
			StringBuilder bitString = new StringBuilder();
			
    		int bits = reader.read(); //read a byte
			while(bits != -1)
	    	{
	    		bitString.append(byteToBitString(bits)); //convert to bitString and add to growing string
	    		bits = reader.read(); //read another byte
	    	}
			reader.close();
			return bitString.toString();
		}
		
		/**
		 * Closes all resources associated with this reader.
		 * @throws IOException
		 */
		public void close() throws IOException
		{
			reader.close();
		}
	}
	
	/**
	 * A class for writing bitstrings to a file (as bits)
	 */
	public static class Writer
	{
		private BufferedWriter writer;
		
		/**
		 * Creates a new Writer for the given file destination
		 * @param destination The file to write to
		 */
		public Writer(File destination) throws IOException
		{
	    	writer = new BufferedWriter(new FileWriter(destination));
		}
		
		/**
		 * Writes a the given text to the destination file. Note that this method adds an additional newline character after the given text.
		 * @param text The text to write to the file
		 * @throws IOException
		 */
		public void writeLine(String text) throws IOException
		{
			writer.write(text);
			writer.newLine();
		}
		
		/**
		 * Writes the given bitString to the destination file (as a sequence of bits).
		 * @param bitString A string representation of the bits to write to the file
		 * @throws IOException
		 */
		public void writeBitString(String bitString) throws IOException
		{
			String byteString;
			
			
			for(int i=0; i<bitString.length(); i+=7)
			{
				if(i+7 < bitString.length()) //if we can grab a full byte
					byteString = bitString.substring(i, i+7);
				else
				{
					byteString = bitString.substring(i); //grab what we can
					int bitsLeft = 7 - byteString.length();
					for(int j=bitsLeft; j>0; j--) //pad out the remaining bits
						byteString += "0";
				}
				writer.write(bitStringToByte(byteString));
			}
			
	    	writer.close();
		}
		
		/**
		 * Closes all resources associated with this writer.
		 * @throws IOException
		 */
		public void close() throws IOException
		{
			writer.close();
		}
	}
	
	
	/**
	 * A static method that converts a bit string to a byte.
	 * Note that to avoid errors on some systems, only the lower 7 bits of a byte are used.
	 *
	 * @param bitString The string of 7 bits (e.g., "0010011") to convert to a byte
	 * @return A byte (a binary number) corresponding to the given bit string
	 */
	public static int bitStringToByte(String bitString)
	{
		if(bitString.length() != 7)
			throw new IllegalArgumentException("Illegal bitstring: "+bitString+"; must have 7 bits to convert to a byte!");

		int bits = 0; //the byte to return--start empty
		for(int i=0; i<7; i++)
		{
			bits = bits << 1; //shift over 1 to make room--this adds a "0" in the last spot

			char c = bitString.charAt(i);
			if(c == '1') //if bit is a 1
				bits = bits + 1;  //add 1, which replaces that 0 with a 1. Otherwise it was correct, so leave it
		}
		return bits; //return the byte
	}

	/**
	 * A static method that converts a byte into a bit string representation
	 * Note that to avoid errors on some systems, only the lower 7 bits of the byte are used.
	 *
	 * @param bits A byte (a binary number)
	 * @return A bit string corresponding to the given byte
	 */
	public static String byteToBitString(int bits)
	{
		StringBuilder bitString = new StringBuilder();
		bits = bits << 1; //throw away leading 0
		for(int i=0; i<7; i++) //read the remaining 7 bits
		{
			int bit = (bits & 128); //this gets the "first" bit (using a "bitwise and")
			if(bit == 0) //if we yanked off nothing
				bitString.append('0');
			else //otherwise we got a 1
				bitString.append('1');
			bits = bits << 1; //shift everyone else over (so can get next piece)
		}
		return bitString.toString();
	}	
}
